#include "../slim.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "../common/objects.h"
#include "../common/scene.h"
#include "./obj_parser.h"
#include "../common/debug.h"
#include "../shader/shaders.h"
#include "../intersect/sphere.h"
#include "../intersect/cylinder.h"
#include "../intersect/plane.h"
#include "../intersect/disc.h"
#include "../intersect/rectangle.h"
#include "../intersect/triangle.h"
#include "../shader/lighting.h"
#include "../common/string_extra.h"
#include "../common/vector.h"
#include "obj_slim_converter.h"
#include "parser.h"

void assign_obj_color(color *c, double o[3])
{
	c->r = o[0] * 255;
	c->g = o[1] * 255;
	c->b = o[2] * 255;
}

void obj_vector_to_slim(vector *slm, obj_vector *obj)
{
	slm->x = obj->e[0];
	slm->y = obj->e[1];
	slm->z = obj->e[2];
}

void add_obj_material_to_object( object *new_object, obj_material **mat_list, int mat_index)
{
	new_object->shader = (void* (*)(void*, color*))get_shader_ptr("general");
	//new_object->shader = (void* (*)(void*, color*))get_shader_ptr("shadow_cost");
	set_misc(new_object, 0, 0, 0, 0, 100);
	make_color(&new_object->amb, 150, 150, 150);
	make_color(&new_object->diff, 150, 150, 150);
	make_color(&new_object->spec, 150, 150, 150);
	
	if(mat_index == -1)
		return;

	
	assign_obj_color(&new_object->amb, mat_list[mat_index]->amb);
	assign_obj_color(&new_object->diff, mat_list[mat_index]->diff);
	assign_obj_color(&new_object->spec, mat_list[mat_index]->spec);
	
	//assuming 1000 sharpness is mirror
	new_object->glossy = 1.0f - mat_list[mat_index]->glossy/1000.f;
	new_object->reflect = mat_list[mat_index]->reflect;
	//new_object->refract = mat_list[mat_index]->refract_index;
	new_object->refract_index = mat_list[mat_index]->refract_index;
	new_object->spec_alpha = mat_list[mat_index]->shiny;
	new_object->trans = 1.0f - mat_list[mat_index]->trans;
	
	print_color(MESSAGE, "amb: ", &new_object->amb);
	print_color(MESSAGE, "diff: ", &new_object->diff);
	print_color(MESSAGE, "spec: ", &new_object->spec);
}

void convert_obj_triangle_to_slim(scene_data *scene, obj_scene_data *data, obj_face *oo)
{
	vector vertices[3];
	object *o = triangle_create(scene->num_objects);
	add_obj_material_to_object(o, data->material_list, oo->material_index);

	for(int i=0; i<3; i++)
	{
		int vector_index = oo->vertex_index[i];
		obj_vector *v = data->vertex_list[vector_index];
		vertices[i].x = v->e[0];
		vertices[i].y = v->e[1];
		vertices[i].z = v->e[2];
	}
	
	triangle_from_points_v(o, (vector*)vertices);
	scene_add_object(scene, o);
	
	print_object(MESSAGE, "added obj triangle", o);
}

void convert_obj_camera_to_slim(scene_data *scene, obj_scene_data *data, obj_camera *camera)
{
	if(camera == NULL)
		return;
	
	obj_vector_to_slim(&scene->camera->pos, data->vertex_list[camera->camera_pos_index]);
	obj_vector_to_slim(&scene->camera->lookP, data->vertex_list[camera->camera_look_point_index]);
	obj_vector_to_slim(&scene->camera->up, data->vertex_normal_list[camera->camera_up_norm_index]);
	make_vector(&scene->camera->look, &scene->camera->lookP, &scene->camera->pos);
	set_tri(&scene->camera->rot, 0, 0, 0);
	printd(MESSAGE, "setting up camera\n");
	scene->camera->units = 5;
}

void convert_obj_light_point_to_slim(scene_data *scene, obj_scene_data *data, obj_light_point *oo)
{
	object *o = make_object(LIGHT_POINT, 1337);
	scene_add_light(scene, o);
	add_obj_material_to_object(o, data->material_list, oo->material_index);
	obj_vector_to_slim(&o->pos, data->vertex_list[ oo->pos_index ]);
	print_object(MESSAGE, "added obj light point", o);
}

void convert_obj_light_disc_to_slim(scene_data *scene, obj_scene_data *data, obj_light_disc *oo)
{
	vector v;
	object *o = make_object(LIGHT_DISC, 1337);
	scene_add_light(scene, o);
	add_obj_material_to_object(o, data->material_list, oo->material_index);
	obj_vector_to_slim(&o->pos, data->vertex_list[ oo->pos_index ]);
	obj_vector_to_slim(&o->norm, data->vertex_normal_list[ oo->normal_index ] );
	o->radius = vector_length( &o->norm );
	normalize(&o->norm);
	
	print_object(MESSAGE, "added obj light disc", o);
}

void convert_obj_sphere_to_slim(scene_data *scene, obj_scene_data *data, obj_sphere *oo)
{
	object *o = sphere_create();
	scene_add_object(scene, o);
	add_obj_material_to_object(o, data->material_list, oo->material_index);
	
	obj_vector_to_slim(&o->pos, data->vertex_list[ oo->pos_index ]);
	obj_vector_to_slim(&o->up, data->vertex_normal_list[ oo->up_normal_index ] );
	obj_vector_to_slim(&o->norm, data->vertex_normal_list[ oo->up_normal_index ] );
	o->radius = vector_length( &o->norm );
	
	print_object(MESSAGE, "added obj sphere", o);
}

int obj_reader_for_slim(scene_data *scene, char *filename)
{
	obj_scene_data data;
	if( !parse_obj_scene(&data, filename) )
		return 0;
	
	if(data.camera != NULL)
		convert_obj_camera_to_slim(scene, &data, data.camera);
	
	for(int i=0; i< data.face_count; i++)
	{
		if(data.face_list[i]->vertex_count == 3)
			convert_obj_triangle_to_slim(scene, &data, data.face_list[i]); 
	}
	for(int i=0; i< data.sphere_count; i++)
		convert_obj_sphere_to_slim(scene, &data, data.sphere_list[i]); 
		
	for(int i=0; i< data.light_disc_count; i++)
		convert_obj_light_disc_to_slim(scene, &data, data.light_disc_list[i]);
		
	for(int i=0; i< data.light_point_count; i++)
		convert_obj_light_point_to_slim(scene, &data, data.light_point_list[i]);
		
	delete_obj_data(&data);
		
	return 1;
}
