#include <time.h>
#include <stdlib.h>
#include "../common/debug.h"
#include "../common/colors.h"
#include "../common/buffer.h"
#include "super_sampler.h"
#include "ray_common.h"
#include "tracer.h"
#include "../shader/post_shaders.h"
#include "../common/misc.h"
#include "../common/vector.h"
#include "../misc/save_frame.h"

extern scene_data* main_scene;
extern void (*update_frame_output)(scene_data *scene);

#define NUM_SAMPLES 9
#define sampler sample_row_jitter
#define get_filter_weight gaussian_filter

void sample_row_random(int j, list *row)
{
	int i, k;
	intersect_data *id;
	color c;
	double new_x, new_y;
	
	for(i=0; i<main_scene->width; i++)
	{
		id = (intersect_data*)get_idata(i,j);
		for(k=0; k<NUM_SAMPLES; k++)
		{
			new_x = (double)i + ((double)rand()/RAND_MAX*1.0 - 0.5);
			new_y = (double)j + ((double)rand()/RAND_MAX*1.0 - 0.5);
			calc_ray(new_x, new_y);
			id->obj->shader(id, &c);
			sample_data *sample = (sample_data*)malloc(sizeof(sample_data));
			color_copy(&sample->c, &c);
			sample->pos.x = new_x;
			sample->pos.y = new_y;
			sample->pos.z = 0;
			list_add_item(&row[i], sample, NULL);
			clear_color(&c);
		}
	}
}

void sample_row_jitter(int j, list *row)
{
	int i, k, m;
	intersect_data *id;
	color c;
	double new_x, new_y;
	
	for(i=0; i<main_scene->width; i++)
	{
		id = (intersect_data*)get_idata(i,j);
		for(k=-1; k<2; k++)
			for(m=-1; m<2; m++)
			{
				new_x = i+m*0.3333 + ((double)rand()/RAND_MAX*0.3333-0.15);
				new_y = j+k*0.3333 + ((double)rand()/RAND_MAX*0.3333-0.15);
				calc_ray(new_x, new_y);
				id->obj->shader(id, &c);
				sample_data *sample = (sample_data*)malloc(sizeof(sample_data));
				color_copy(&sample->c, &c);
				sample->pos.x = new_x;
				sample->pos.y = new_y;
				sample->pos.z = 0;
				list_add_item(&row[i], sample, NULL);
				clear_color(&c);
			}
	}
}

void sample_row_uniform(int j, list *row)
{
	int i, k, m;
	intersect_data *id;
	color c;
	double new_x, new_y;
	
	for(i=0; i<main_scene->width; i++)
	{
		id = (intersect_data*)get_idata(i,j);
		for(k=-1; k<2; k++)
			for(m=-1; m<2; m++)
			{
				new_x = i+m*0.3333;
				new_y = j+k*0.3333;
				calc_ray(new_x, new_y);
				id->obj->shader(id, &c);
				sample_data *sample = (sample_data*)malloc(sizeof(sample_data));
				color_copy(&sample->c, &c);
				sample->pos.x = new_x;
				sample->pos.y = new_y;
				sample->pos.z = 0;
				list_add_item(&row[i], sample, NULL);
			}
	}
}

double box_filter(double xdis, double ydis, double r)
{
	if(xdis < r && ydis < r)
		return 1.0;
	return 0.0;
}

double triangle_filter(double xdis, double ydis, double r)
{
	double dis = sqrt(xdis*xdis + ydis*ydis);
	if(xdis < r && ydis < r)
		return 1.0 - dis/r;
	return 0.0;
}

double quadratic_filter(double xdis, double ydis, double r)
{
	double dis = sqrt(xdis*xdis + ydis*ydis);
	if(xdis < r && ydis < r)
		return 1.0 - pow(dis,2)/pow(r,2);
	return 0.0;
}

double gaussian_filter(double xdis, double ydis, double r)
{
	double dis = sqrt(xdis*xdis + ydis*ydis);
	double e = exp(- pow(dis,2)/(2*pow(r,2)) );
	double c = 1/(sqrt(2*M_PI)*r);
	return c*e;
}

void reconstruct(list **row, int row_index)
{
	int col_index, sample_index, offset_col, offset_row, offset;
	color accum, c;
	sample_data *sample;
	double scale;
	double xdis, ydis, weight;
	point center;
	double filter_size = 0.6;
			
	for(col_index=1; col_index<main_scene->width-1; col_index++)
	{
		center.x = col_index;
		center.y = row_index;
		center.z = 0.0;
		
		scale = 0.0;
		clear_color(&accum);
		
		for(offset_row=0; offset_row<3; offset_row++)
		{
			for(offset=-1; offset<2; offset++)
			{
				offset_col = col_index+offset;
				for(sample_index=0; sample_index<NUM_SAMPLES; sample_index++)
				{
					sample = (sample_data*)list_get_index(&row[offset_row][offset_col], sample_index);
					color_copy(&c, &sample->c);
					xdis = fabs(sample->pos.x - center.x);
					ydis = fabs(sample->pos.y - center.y);
					weight = get_filter_weight(xdis, ydis, filter_size);
					multiply_color(&c, weight);
					scale += weight;
					add_colors(&accum, &c);
				}
			}
		}
		
		multiply_color(&accum, (double)1.0/scale);
		set_color(col_index, row_index, &accum);
	}
}

void delete_row_samples(list *row)
{
	int i, k;
	for(i=0; i<main_scene->width; i++)
	{
		for(k=0; k<NUM_SAMPLES; k++)
		{
			sample_data *s = (sample_data*)list_get_index(&row[i], k);
			free( s );
		}
		
		list_delete_all(&row[i]);
	}
}

void super_sampler_render()
{
	int i, j;
	list **row;
	list *temp_list;
	
	row = (list**) malloc(sizeof(list*)*3);
	row[0] = (list*) malloc(sizeof(list)*main_scene->width);
	row[1] = (list*) malloc(sizeof(list)*main_scene->width);
	row[2] = (list*) malloc(sizeof(list)*main_scene->width);
	
	for(i=0; i<main_scene->width; i++)
	{
		list_make(&row[0][i], NUM_SAMPLES, 1);
		list_make(&row[1][i], NUM_SAMPLES, 1);
		list_make(&row[2][i], NUM_SAMPLES, 1);
	}

	j = 0;
	sampler(j, row[0]);
	j = 1;
	sampler(j, row[1]);
	j = 2;
	sampler(j, row[2]);
	
	for(j=1; j<main_scene->height-1; j++)
	{
		reconstruct(row, j);
		
		if(j % 10 == 0)
			update_frame_output(main_scene);

		delete_row_samples(row[0]);
		

		temp_list = row[0];
		row[0] = row[1];
		row[1] = row[2];
		row[2] = temp_list;

		sampler(j, row[2]);
	}
}
