#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "../common/objects.h"
#include "../common/vector.h"
#include "../intersect/intersect.h"
#include "../common/structs.h"
#include "../common/debug.h"
#include "phong.h"
#include "lighting.h"

#include "../common/colors.h"

extern scene_data* main_scene;

int shadow_cache_id = -1;
int shadow_cache_max_misses = 5;
int shadow_cache_misses = 0;
#define USE_SHADOW_CACHE
//#define DEBUG_CACHE

//double shadow_cache_test(vector *origin, vector *light_dir, double distance_to_light)
double shadow_cache_test(vector *origin, vector *light_dir, double distance_to_light, object *testobj)
{
	double distance_closest_object;
	vector new_intersection;
	object *obj_hit;
	
	// cached object test
#ifdef USE_SHADOW_CACHE
	if(shadow_cache_id > 0)
	{
		distance_closest_object = intersect_single_object(origin, light_dir, main_scene->models[shadow_cache_id], 1);
		if(distance_closest_object > EPSILON && distance_closest_object < distance_to_light)
		{
			shadow_cache_misses = 0;
			#ifdef DEBUG_CACHE
			color_set_rgb(&testobj->amb, 0, 255, 0);
			color_set_rgb(&testobj->diff, 0, 255, 0);
			#endif
			return distance_closest_object;
		}
	}
#endif
	
	// all object test
	obj_hit = intersect_all_objects_getptr(origin, light_dir, &new_intersection, 1);
	distance_closest_object = distance_between(origin, &new_intersection);
	if ( distance_closest_object < distance_to_light )
	{
		shadow_cache_id = obj_hit->id;
		shadow_cache_misses = 0;
		#ifdef DEBUG_CACHE
		color_set_rgb(&testobj->amb, 0, 0, 255);
		color_set_rgb(&testobj->diff, 0, 0, 255);
		#endif
		return distance_closest_object;
	}

	// no shadow
	
	if(shadow_cache_misses > shadow_cache_max_misses)
	{
		shadow_cache_id = -1;
		#ifdef DEBUG_CACHE
		color_set_rgb(&testobj->amb, 255, 0, 0);
		color_set_rgb(&testobj->diff, 255, 0, 0);
		#endif
	}
	else
	{
		shadow_cache_misses++;
		#ifdef DEBUG_CACHE
		color_set_rgb(&testobj->amb, 255, 255, 0);
		color_set_rgb(&testobj->diff, 255, 255, 0);
		#endif
	}

	return 0.0f;
}


void light_disc_shadow(intersect_data *id, color *shadowed_color)
{
	vector light_dir;
	vector light_pos;
	vector random_v;
	vector *hit_point = &id->intersect;
	int rays_tried = 0;
	int visible_rays = 0;
	double radius = main_scene->lights[id->other_obj_num]->radius;
	double last_shadow_percent = 1.0;
	double allowed_variance = SHADOW_SAMPLE_VARIANCE;
	int check_every = SHADOW_CHECK_EVERY;
	int next_check = check_every;
	double distance_to_light;
	double distance_closest_object;
	
	for(rays_tried = 1; rays_tried <= SHADOW_RAY_SAMPLES; rays_tried++)
	{
		//random point on light surface
		random_vector_unitsphere(&random_v);
		cross_product(&light_pos, &random_v, &main_scene->lights[id->other_obj_num]->norm);
		normalize(&light_pos);  //some vector in disc plane now

		//lastly, vector needs a random length
		multiply_vector(&light_pos, &light_pos, radius);
		//vector_scalar_subtract(&light_pos, radius);

		add_vectors(&light_pos, &light_pos, &main_scene->lights[id->other_obj_num]->pos);

		//from the object to the light
		make_vector(&light_dir, &light_pos, hit_point);
		normalize(&light_dir);


		distance_to_light = distance_between(hit_point, &light_pos);
		distance_closest_object = shadow_cache_test(hit_point, &light_dir, distance_to_light, id->obj);
		if(distance_closest_object == 0.0f)
		{
			visible_rays++;
		}
		
shadowed_color->g = (double)rays_tried;
		
		//shadow adaptively
		//keep going only if the shadow is getting darker
		if(rays_tried == next_check)
		{
			shadowed_color->r = (double)visible_rays / (double)rays_tried;
			if(last_shadow_percent == shadowed_color->r)
			{
				return;
			}
			double diff = fabs(last_shadow_percent - shadowed_color->r);
			if(diff < allowed_variance)
			{
				return;
			}
			last_shadow_percent = 	shadowed_color->r;
			next_check += check_every;
		}
	}

	shadowed_color->r = (double)visible_rays / (double)(rays_tried-1);
}


void light_sphere_shadow(intersect_data *id, color *shadowed_color)
{
	vector light_dir;
	point light_pos;
	point shadow_hit;
	point *hit_point = &id->intersect;
	int rays_tried = 0;
	int visible_rays = 0;
	double two_radius = main_scene->lights[id->other_obj_num]->radius * 2;
	double radius = main_scene->lights[id->other_obj_num]->radius;
	double last_shadow_percent = 1.0;
	double allowed_variance = 0.0;
	int check_every = 100;
	int next_check = check_every;
	double distance_to_light;
	double distance_closest_object;
	
	for(rays_tried = 1; rays_tried <= SHADOW_RAY_SAMPLES; rays_tried++)
	{
		//random point on light surface
		random_vector(&light_pos);
		multiply_vector(&light_pos, &light_pos, two_radius);
		vector_scalar_subtract(&light_pos, radius);
		add_vectors(&light_pos, &light_pos, &main_scene->lights[id->other_obj_num]->pos);

		//from the object to the light
		make_vector(&light_dir, &light_pos, hit_point);
		normalize(&light_dir);


		distance_to_light = distance_between(hit_point, &light_pos);
		distance_closest_object = shadow_cache_test(hit_point, &light_dir, distance_to_light, id->obj);
		if(distance_closest_object == 0.0f)
		{
			visible_rays++;
		}
		
shadowed_color->g = (double)rays_tried;
		
		//shadow adaptively
		//keep going only if the shadow is getting darker
		if(rays_tried == next_check)
		{
			shadowed_color->r = (double)visible_rays / (double)rays_tried;
			if(last_shadow_percent == shadowed_color->r)
			{
				return;
			}

			if(fabs(last_shadow_percent - shadowed_color->r) < allowed_variance)
			{
				return;
			}
			last_shadow_percent = 	shadowed_color->r;
			next_check += check_every;
		}
	}

	shadowed_color->r = (double)visible_rays / (double)(rays_tried-1);
}

void light_point_shadow(intersect_data *id, color *shadowed_color)
{
	vector *origin = &id->intersect;
	vector light_dir;
	double distance_to_light;
	double distance_closest_object;
	
	sub_vectors(&light_dir, &main_scene->lights[id->other_obj_num]->pos, origin);
	normalize(&light_dir);
	
	distance_to_light = distance_between(origin, &main_scene->lights[id->other_obj_num]->pos);
	distance_closest_object = shadow_cache_test(origin, &light_dir, distance_to_light, id->obj);
	if(distance_closest_object > 0.0f)
	{
		shadowed_color->r = 0.0f;
		return;
	}

	shadowed_color->r = 1.0f;
	return;
}



void light_infinite_shadow(intersect_data *id, color *shadowed_color)
{
//	object *ob;
//	point tracePt;	//1st object on ray from object to light
//	point *intersectP = &id->intersect;

	//light is at infinity, so the pos is the light vector
	//the position should already be normalized

	// see if there is an object from intersection to light
	//ob = intersect(intersectP, &main_scene->lights[light_num]->pos, &tracePt, 1);

	//if there is any object, it block the light
	//if ( ob != NULL ) 
	//	return 0;
	shadowed_color->r = 1;
	return;
}


void light_spotlight_shadow(intersect_data *id, color *shadowed_color)
{
	vector lightV;
	point lightP;
	object *ob;
	vector spotV;
	point tracePt;	//1st object on ray from object to light
	point *intersectP = &id->intersect;
	//double dis = 0.5;
	//int i;
	double LdpL;


	/* light vector */
	lightV.x = main_scene->lights[id->other_obj_num]->pos.x - intersectP->x;
	lightV.y = main_scene->lights[id->other_obj_num]->pos.y - intersectP->y;
	lightV.z = main_scene->lights[id->other_obj_num]->pos.z - intersectP->z;
	normalize(&lightV);
	
	make_vector(&spotV, &main_scene->lights[id->other_obj_num]->pos, &main_scene->lights[id->other_obj_num]->norm);
	normalize(&spotV);
	LdpL = dot_product(&lightV, &spotV);

	if( LdpL <= 0 || LdpL < 0 )
	{
		shadowed_color->r = 0;
		return;
	}
	
	//scale
	LdpL = 1 - distance_between(&spotV, &lightV) / main_scene->lights[id->other_obj_num]->radius;
	if(LdpL > 1)
	{
		shadowed_color->r = 0;
		return;
	}

	LdpL = LdpL * main_scene->lights[id->other_obj_num]->radius;

	lightP.x = main_scene->lights[id->other_obj_num]->pos.x;
	lightP.y = main_scene->lights[id->other_obj_num]->pos.y;
	lightP.z = main_scene->lights[id->other_obj_num]->pos.z;


	/* calculate shadow */
	ob = intersect_all_objects_getptr(intersectP, &lightV, &tracePt, 1);

	/* if we are in a shadow, try the next light (ambient is the only
	 * lighting needed and it is already done */
	if ( ob != NULL && distance_between(intersectP, &tracePt) <
			distance_between(intersectP, &lightP))
		shadowed_color->r = 0;
	else
		shadowed_color->r = LdpL * 1;

	return;
}


void get_shadow(intersect_data *id, color *shadowed_color)
{
	shadowed_color->r = 1.0f;

	if (main_scene->lights[id->other_obj_num]->obj_type == LIGHT_INFINITE)
		//shadow = pointlight(light_num, id);
		//shadow = 1;
		return;

	else if( main_scene->lights[id->other_obj_num]->obj_type == LIGHT_SPOT)
		light_spotlight_shadow(id, shadowed_color);

	else if( main_scene->lights[id->other_obj_num]->obj_type == LIGHT_POINT)
		light_point_shadow(id, shadowed_color);

	else if (main_scene->lights[id->other_obj_num]->obj_type == LIGHT_SPHERE)
		light_sphere_shadow(id, shadowed_color);
	
	else if (main_scene->lights[id->other_obj_num]->obj_type == LIGHT_DISC)
		light_disc_shadow(id, shadowed_color);
#ifdef DEBUG_CACHE
shadowed_color->r = 1.0f;
#endif
//	return 1.0f;
	//return shadow;
}
