#include "image_source.h"
#include "../common/objects.h"
#include "../common/scene.h"
#include "../intersect/sphere.h"
#include "../intersect/line_segment.h"
#include "../common/colors.h"
#include "../intersect/intersect.h"
#include "../parse/parser.h"

#include <string.h>

#define SAMPLE_RATE 44100
#define IR_LENGTH 10
#define REFL_COEFFICIENT 0.7
#define SPEED_SOUND 343.0
#define tree_object(tree, order) (main_scene->models[tree->object_list[order]])


extern scene_data* main_scene;
extern color white;
extern color red;
extern color green;
extern color blue;
extern color yellow;
extern color purple;
object *source;
object *listener;

double impulse_response[IR_LENGTH * SAMPLE_RATE];
char visualize;

FILE *general_out_file;
FILE *matlab_out_file;

///////////////////////////
// little helper functions
///////////////////////////

void store_ir(reflection_tree *tree)
{
	double time;
	double attenuation;
	double distance;
	
	distance = distance_between(&tree->image_list[tree->order], &listener->pos);
	time = SAMPLE_RATE * (distance / SPEED_SOUND);
	
	attenuation = pow(REFL_COEFFICIENT, tree->order);
	if(distance < 1)
		distance  = 1;
	attenuation = attenuation * (1 / distance);
	
	impulse_response[(int)time] += attenuation;
}

void add_sphere_source(vector *pos, color *c)
{
	object *obj = sphere_create();
	
	
	copy_vector(&obj->pos, pos);
	obj->radius = 0.5 + 0.1 * (double)rand()/RAND_MAX;
	
	//add sphere to scene models
	obj->id = main_scene->num_objects;
	main_scene->models[main_scene->num_objects] = obj;
	main_scene->num_objects++;
	
	color_copy(&obj->diff, c);
	color_copy(&obj->amb, c);
	color_copy(&obj->spec, &white);
}


void add_line_path(vector *start, vector *end, color *c)
{
	object *obj = line_segment_create(4);
	
	if( !visualize)
		return;
	
	//add_sphere_source(start, &red);
	//add_sphere_source(end, &red);
	
	copy_vector(&obj->pos, start);
	copy_vector(&obj->norm, end);
	
	line_segment_setup(obj);
	scene_add_object(main_scene, obj);
	
	color_copy(&obj->diff, c);
	color_copy(&obj->amb, c);
	color_copy(&obj->spec, c);
}

void clear_image_structure(reflection_tree *tree)
{
	int i;
	
	for(i=0; i<IR_LENGTH*SAMPLE_RATE; i++)
		impulse_response[i] = 0;
	
	for(i=0; i<MAX_ORDER; i++)
	{
		tree->contributions[i] = 0;
		tree->object_list[i] = NULL;
	}
	
	tree->order = 0;
}



void pick_first_two_spheres()
{
	int i;
	int j;
	
	for(i=0; i<main_scene->num_objects; i++)
	{
		if( main_scene->models[i]->obj_type != SPHERE)
			continue;
		
		source = main_scene->models[i];
		for(j=i+1; i<main_scene->num_objects; j++)
		{
			if( main_scene->models[j]->obj_type != SPHERE)
				continue;
			
			listener = main_scene->models[j];
			break;
		}
		break;
	}
	
	for(i=0; i<main_scene->num_objects; i++)
	{
		if( main_scene->models[i]->obj_type == RECTANGLE)
		{
			make_color(&main_scene->models[i]->amb, 150, 250, 200);
			make_color(&main_scene->models[i]->diff, 150, 250, 120);
			main_scene->models[i]->trans = 0.7;
		}
	}
}


double wall_intersect(point* startP, vector* projV, object *skip_wall)
{
	double distanceOld =  INFINITY-1;
	double distanceNew = INFINITY;
	int i;
	
	for(i=0; main_scene->models[i] != NULL; i++)
	{
		if(main_scene->models[i]->obj_type != RECTANGLE)
			continue;
	
		if( main_scene->models[i]==skip_wall )
			continue;
		
		distanceNew = intersect_single_object(startP, projV, main_scene->models[i], 0);
		
		if(distanceNew> EPSILON)
		{
			if(distanceNew < distanceOld + EPSILON)
			{
				distanceOld = distanceNew;
			}
		}
	}
	
	return distanceOld;
}

///////////////////////////
// image source functions
///////////////////////////

char can_reflect_point(vector *start, object *rect, vector *reflection)
{
	vector start_to_rect;
	vector direction;
	double distance_between;
	
	sub_vectors(&start_to_rect, &rect->pos,  start);
	
	//check if normal is correct
	if(dot_product(&start_to_rect, &rect->norm) > 0)
		return 0;
	
	distance_between = dot_product(&start_to_rect, &rect->norm);
	
	multiply_vector(&direction, &rect->norm, 2*distance_between);
	add_vectors(reflection, start, &direction);
	
	return 1;
}


char obstructed(vector *original_start, vector *direction, vector *disired_end, object *skip_wall)
{
	vector start;
	vector to_disired_end;
	double distance;
	double new_distance;

	copy_vector(&start, original_start);
	//end point must be in front!
	sub_vectors(&to_disired_end, disired_end, original_start);
	if( vector_is_opposite(direction, &to_disired_end) )
		return 1;
	
	//move the start a tiny bit
//	multiply_vector(&tiny_bit, direction, EPSILON);
//	add_vectors(&start, original_start, &tiny_bit);
	
	distance = distance_between(&start, disired_end);
	new_distance = wall_intersect(&start, direction, skip_wall);
	
	if(new_distance > EPSILON)
	{
		//freakish corner case
		if(new_distance == distance)
			return 0;
		if(new_distance < distance)
			return 1;
	}

	return 0;
}

char store_tree_reflection(object *rect, reflection_tree *tree)
{
	if( !can_reflect_point(&tree->image_list[tree->order-1], rect, &tree->image_list[tree->order]) )
		return 0;

	tree->object_list[tree->order] = rect->id;
	
	return 1;
}

char passes_through(vector *from, vector *to, object *through, vector *hit_point, vector *direction)
{
	double distance;
	double max_distance;
	
	sub_vectors(direction, to, from);
	max_distance = vector_length(direction);

	normalize(direction);
	
	distance = through->intersect(from, direction, through, NULL);
	
	if(distance < EPSILON || distance > max_distance - EPSILON)
		return 0;
	
	multiply_vector(hit_point, direction, distance);
	add_vectors(hit_point, hit_point, from);
	return 1;
}

char is_visible(reflection_tree *tree)
{
	vector reflected_listener;
	vector hit_point;
	vector other_hit_point;
	vector dir;
	vector hits[4][2];
	char current_order;
	
	//maximum order test
	{
		current_order = tree->order;
		copy_vector(&reflected_listener, &listener->pos);
		if( !passes_through(&reflected_listener, &tree->image_list[current_order],
								 tree_object(tree, current_order), &hit_point, &dir) )
			return 0;
		
		//internal obstruction test
		//current hit point to previous hitpoint (listener in this case)
		if( obstructed(&reflected_listener, &dir, &hit_point, tree_object(tree, current_order)) )
			return 0;
		
		copy_vector(&hits[current_order][0], &reflected_listener);
		copy_vector(&hits[current_order][1], &hit_point);
	}

	
	for(current_order=tree->order-1; current_order>0; current_order--)
	{
		if( !can_reflect_point(&reflected_listener, tree_object(tree, current_order+1), &reflected_listener) )
			return 0;
		
		if( !passes_through(&reflected_listener, &tree->image_list[current_order],
								 tree_object(tree, current_order), &other_hit_point, &dir) )
			return 0;
		
		if( !passes_through(&tree->image_list[current_order], &reflected_listener,
								 tree_object(tree, current_order+1), &hit_point, &dir) )
			return 0;
		
		//internal obstruction test
		//current hit point to previous hitpoint
		if( obstructed(&other_hit_point, &dir, &hit_point, tree_object(tree, current_order+1)) )
			return 0;

		//check if reflected across invalid walls
		//crosses wall with normal facing towards source
//		if( dot_product(&dir, &tree_object(tree, current_order+1)->norm) < 0)
//			return 0;
		
		copy_vector(&hits[current_order][0], &other_hit_point);
		copy_vector(&hits[current_order][1], &hit_point);
	}
	
	//minimum order test
	{
		current_order = 0;
		if( !can_reflect_point(&reflected_listener, tree_object(tree, current_order+1), &reflected_listener) )
			return 0;
		
		if( !passes_through(&tree->image_list[current_order], &reflected_listener,
								 tree_object(tree, current_order+1), &hit_point, &dir) )
			return 0;
		
		//internal obstruction test
		//current hit point (source in this case) to previous hitpoint
		if( obstructed(&tree->image_list[current_order], &dir, &hit_point, tree_object(tree, current_order+1)) )
			return 0;

		copy_vector(&hits[current_order][0], &tree->image_list[current_order]);
		copy_vector(&hits[current_order][1], &hit_point);
	}
	

	for(current_order=0; current_order<=tree->order; current_order++)
	{
		add_line_path(&hits[current_order][0], &hits[current_order][1], &purple);
	}
	
	return 1;
}


void create_all_reflections(reflection_tree *tree)
{
	int i,j;
//	vector source_to_rect;
	object *rect;
	const char debug_on = 0;
	char hits[3] = {3,5};
	visualize = 1;

	if(tree->order == 0)
	{
		vector dir;
		sub_vectors(&dir, &listener->pos, &source->pos);
		normalize(&dir);
		if( ! obstructed(&source->pos, &dir, &listener->pos, NULL) )
		{
			print_file(NORMAL, general_out_file, "order[0] [] ");
			print_file(NORMAL, general_out_file, "%.3f\n", distance_between(&source->pos, &listener->pos) );
			
			store_ir(tree);
			tree->contributions[tree->order]++;
		}
	}
	
	tree->order++;
	
	for(i=0; main_scene->models[i] != NULL; i++)
	{
		if(debug_on)
			i = hits[tree->order-1];

		
		rect = main_scene->models[i];
		if( rect->obj_type != RECTANGLE)
			if(debug_on) {return;} else {
				continue;}

		//reflect the new image
		if( !store_tree_reflection(rect, tree) )
			if(debug_on) {return;} else {
				continue;}
		
		//check visibility
		if( is_visible(tree) )
		{
			//visualize
			/*
			if(tree->order == 1)
				add_sphere_source(&tree->image_list[tree->order], &red);
			if(tree->order == 2)
			{
				add_sphere_source(&tree->image_list[tree->order-1], &red);
				add_sphere_source(&tree->image_list[tree->order], &green);
			}
			if(tree->order == 3)
			{
				add_sphere_source(&tree->image_list[tree->order-2], &red);
				add_sphere_source(&tree->image_list[tree->order-1], &green);
				add_sphere_source(&tree->image_list[tree->order], &blue);
			}*/
			
			 
			//collect contributions
			store_ir(tree);
			
			tree->contributions[tree->order]++;
			
			print_file(NORMAL, general_out_file, "order[%i] [", tree->order);
			for(j=1; j<tree->order; j++)
				print_file(NORMAL, general_out_file, "%i,", tree->object_list[j]);
			print_file(NORMAL, general_out_file, "%i] ", tree->object_list[tree->order]);
			print_file(NORMAL, general_out_file, "%.3f\n", 
					   distance_between(&tree->image_list[tree->order], &listener->pos) );
		}
		
			
		if(tree->order < main_scene->max_recurs)
			create_all_reflections(tree);
		
		if(debug_on)
			return;
	}
	
	tree->order--;
}


///////////////////////////
// print functions
///////////////////////////

void print_simple_ir(FILE *file)
{
	int i;
	
	print_file(NORMAL, file, "\n");
	for(i=0; i<IR_LENGTH*SAMPLE_RATE; i++)
	{
		if(impulse_response[i] != 0)
			print_file(NORMAL, file, "time: %.3fs atten: %.3f\n", (double)i/SAMPLE_RATE, impulse_response[i]);
	}
}

void print_matlab_simple_ir(FILE *file)
{
	int i;
	int start_zero = 0;
	int end_zero = -1;
	
	print_file(NORMAL, file, "\n");
	for(i=0; i<IR_LENGTH*SAMPLE_RATE; i++)
	{
		if(impulse_response[i] != 0)
		{
			if(end_zero != -1)
				print_file(NORMAL, file, " zeros(1,%i) ", end_zero - start_zero + 1);
			
			print_file(NORMAL, file, "%.5f ", impulse_response[i]);
			start_zero = i+1;
			end_zero = -1;
		}
		else
		{
			end_zero = i;
		}
	}
	
	if(end_zero - start_zero > 0)
		print_file(NORMAL, file, " zeros(1,500)\n");
	
	print_file(NORMAL, file, "\n");
}


void print_matlab_table_ir(FILE *file)
{
	int i;
	
	print_file(NORMAL, file, "\n");
	for(i=0; i<IR_LENGTH*SAMPLE_RATE; i++)
	{
		if(impulse_response[i] != 0)
			print_file(NORMAL, file, "%i %f\n", i+1, impulse_response[i]);
	}
}

void print_result_data(reflection_tree *tree)
{
	int total_contribs = 0;
	
	print_matlab_simple_ir(general_out_file);
	print_simple_ir(general_out_file);
	print_matlab_table_ir(matlab_out_file);
	
	print_file(NORMAL, general_out_file, "\n");
	for(int i=0; i<MAX_ORDER; i++)
	{
		print_file(NORMAL, general_out_file, "%i order contrib: %i\n", i, tree->contributions[i]);
		total_contribs += tree->contributions[i];
	}
	print_file(NORMAL, general_out_file, "total contrib: %i\n", total_contribs);
}

///////////////////////////
// output files
///////////////////////////
void open_files()
{
	char data_name[NAME_LENGTH];
	strncpy(data_name, main_scene->scene_name, NAME_LENGTH);
	strcat(data_name, "-IR_data.txt");
	general_out_file = fopen(data_name,"w");
	
	strncpy(data_name, main_scene->scene_name, NAME_LENGTH);
	strcat(data_name, "-IR.m");
	matlab_out_file = fopen(data_name,"w");
}

void close_files()
{
	fclose(matlab_out_file);
	fclose(general_out_file);
}

///////////////////////////
// entry point to image source method
///////////////////////////
void setup_image_source()
{
	reflection_tree *image_tree;
	
	source = get_object_name("source");
	listener = get_object_name("listener");
	
//	if( strequal(main_scene->extension, "obj") )
		pick_first_two_spheres();
	
	if(source == NULL || listener == NULL)
	{
		printd(ERROR, "Need a source and listener!\n");
		return;
	}
	
	make_color(&source->amb, 250, 250, 250);
	make_color(&source->diff, 250, 250, 250);
	make_color(&listener->amb, 75, 75, 75);
	make_color(&listener->diff, 75, 75, 75);

	open_files();
	
	image_tree = (reflection_tree*) malloc( sizeof(reflection_tree) );
	clear_image_structure(image_tree);
	copy_vector(&image_tree->image_list[0], &source->pos);
	
	create_all_reflections(image_tree);
	
	print_result_data(image_tree);
	
	free(image_tree);
	close_files();
	
	//exit(0);
}
